/*
 * action.c
 *
 * Handle action queue events
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "graphics.h"
#include "grv.h"

/* ------------------------------------------------------------------------ *
 *  Primary action event queue machinery
 * ------------------------------------------------------------------------ */

struct action {
    int x, y;
    enum actions what;
    unsigned int priority;
    double when;
    uint64_t seq;
};

static struct action *actions;
static unsigned int action_count;
static unsigned int max_actions = 0;

/*
 * == actions + action_count - 1 if the list is sorted, otherwise NULL
 */
static const struct action *next_action;

/*
 * Comparison function for addaction() and sort_actions().
 * Returns true if *a > *b.
 * This is somewhat the opposite of the typical, but the list is sorted
 * in reverse order.
 */
static inline bool
actions_compare(const struct action *a, const struct action *b)
{
    /* This *must* be a stable sort! */
    if (a->when != b->when)
	return a->when > b->when;
    if (a->priority != b->priority)
	return a->priority > b->priority;

    /* The sequence number will never duplicate */
    return a->seq > b->seq;
}

/* qsort()-compatible interface */
static int actions_compare_qsort(const void *a, const void *b)
{
    return actions_compare(a, b) ? -1 : 1;
}

/*
 * This sorts the list of actions in reverse order of time
 */
static void sort_actions(void)
{
    if (next_action)
	return;			/* Already sorted */

    qsort(actions, action_count, sizeof actions[0], actions_compare_qsort);
    next_action = actions + (action_count - 1);	/* Now sorted */
}

/*
 * Insert an action into an already sorted list; used for events
 * added while the game is already in progress.
 */
static void insert_action(const struct action *act)
{
    size_t a = 0;
    size_t b = action_count;

    while (a < b) {
	size_t c = (a + b) >> 1;
	if (actions_compare(&actions[c], act))
	    a = c + 1;
	else
	    b = c;
    }

    if (a != action_count)
	memmove(&actions[a + 1], &actions[a],
		(action_count - a) * sizeof(actions[0]));

    actions[a] = *act;
    action_count++;
    next_action++;
}

void addaction(int x, int y, double when, unsigned int priority,
	       enum actions what)
{
    static uint64_t next_seq;
    struct action act;

    if (action_count >= max_actions) {
	/* It is assumed unlikely that the number below will ever be exceeded... */
	max_actions += max_actions + 256;
	actions = realloc(actions, max_actions * sizeof(*actions));
#ifdef HAVE_MALLOC_USABLE_SIZE
	max_actions = malloc_usable_size(actions) / sizeof(*actions);
#endif
	if (next_action)
	    next_action = actions + action_count - 1;
    }

    act.x = x;
    act.y = y;
    act.when = when;
    act.priority = priority;
    act.what = what;
    act.seq = ++next_seq;

    if (next_action) {
	/* List already sorted, e.g. game time - simple insertion */
	insert_action(&act);
    } else {
	/* Just stash it at the end */
	actions[action_count++] = act;
    }
}

/* Remove any action which matches the given (x,y,what) */
void removeaction(int x, int y, enum actions what)
{
    size_t i;

    for (i = 0; i < action_count; i++) {
	if (actions[i].x == x && actions[i].y == y && actions[i].what == what) {
	    actions[i] = actions[--action_count];
	    next_action = NULL;	/* List is now unsorted */
	}
    }
}

void reset_actions(void)
{
    action_count = 0;
    next_action = NULL;		/* Many adds are likely to happen next */
}

/* This changes the time of *all* actions of a specifc type to "when" */
void retime_all(double when, enum actions what)
{
    unsigned int i;

    for (i = 0; i < action_count; i++) {
	if (actions[i].what == what) {
	    actions[i].when = when;
	    next_action = NULL;	/* List is now unsorted */
	}
    }
}

/* ------------------------------------------------------------------------ *
 *  Stuff that gets triggered by action events...
 * ------------------------------------------------------------------------ */

/* Blow bomb or smash */
static void blow_bomb(int x, int y, int smash)
{
    /* The first parameter is how far above and below (x,y) does this happen? */
    static const int bomb_pattern[] = {
	5, 0, 0, 2, 2, 3, 3, 3, 2, 2, 0, 0
    };
    static const int smash_pattern[] = {
	6, 0, 2, 3, 4, 4, 5, 5, 5, 4, 4, 3, 2, 0
    };
    static const int big_smash_pattern[] = {
	9, 0, 2, 4, 5, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 5, 4, 2, 0
    };
    const int *pattern;
    int path, patw;
    int x2, y2;
    int ch, m, fl;
    int erase;

    if (!smash)
	gp.Bombs--;

    pattern = smash ? (rnd() < 0.1 ? big_smash_pattern : smash_pattern)
	: bomb_pattern;

    patw = *pattern++;

    /* Note the order in which this happens.  We escavate from the bottom up
       in each column; this means that when we run into rocks we can just let
       them fall.  We can also activate rocks and apples above the hole
       at the end of each x2 loop.

       This was a bug in the original BASIC version... it didn't do this. */

    for (y2 = y - patw; y2 <= y + patw; y2++) {
	path = *pattern++;

	if (y2 < 1 || y2 > 40)
	    continue;

	for (x2 = x + path; x2 >= x - path; x2--) {
	    if (x2 < 2 || x2 > 23)
		continue;

	    ch = screen(x2, y2);
	    m = fg(x2, y2);
	    fl = screen1(x2, y2) & 0x80;	/* Flashing? */
	    erase = 1;

	    switch (ch) {
	    case SYM_PLAYER:
		if (smash)
		    erase = 0;
		else
		    gp.Status = Status_Dead;
		break;

	    case SYM_GHOST:
		kill_ghost(x2, y2);
		break;

	    case SYM_CHERRY:
		take_cherry();
		break;

	    case SYM_BONUS:
		if (gp.Bon)
		    gp.Sc += 15 * gp.Level;	/* This sucks :) */
		break;

	    case SYM_APPLE:
		erase = 0;
		color(m + 16, 0);
		lprint(x2, y2, "\xfe");
		if (!fl)
		    addaction(x2, y2, gp.Tid + 30.0 / (gp.Level + 1), -1,
			      act_apple);
		break;

	    case SYM_BOMB:
	    case_door:
	    case_hyper:
		/* Don't erase, but put it on dark background */
		if (x2 != x || y2 != y) {
		    erase = 0;
		    if (m == 0)
			m = bg(x2, y2);
		    color(m + (fl ? 16 : 0), 0);
		    lprintf(x2, y2, "%c", ch);
		}
		break;

	    case SYM_ROCK:
		fall_rock(x2, y2);
		erase = 0;
		break;

	    default:
		break;
	    }

	    if (erase) {
		color(m, 0);
		lprint(x2, y2, " ");
	    }
	}

	/* See if there is something above this row that should be activated */
	check_activation(x2, y2);
    }

    update_score();
}

/* Turn on or off the bonus dots */
static void bonus_on(void)
{
    gp.Bon = 1;
    color(30, 0);
    lprint(25, 1, " BONUS ");
    addaction(0, 0, gp.Tid + 12.0, -1, act_bonus_off);
}

static void bonus_off(void)
{
    gp.Bon = 0;
    color(0, gp.c);
    lprint(25, 1, "       ");
    addaction(0, 0, gp.Tid + 120.0, 0, act_bonus_on);
}

/* ------------------------------------------------------------------------ *
 *  Action event queue trigger function
 * ------------------------------------------------------------------------ */

void handle_action(void)
{
    const struct action *act;

    if (!action_count)
	return;			/* Queue is totally empty */

    if (!next_action) {
	/* Not sorted yet... */
	sort_actions();
    }

    act = next_action;
    if (act->when > gp.Tid)	/* Not yet */
	return;

    /* Remove from list */
    next_action--;
    action_count--;

    switch (act->what) {
    case act_none:
	break;

    case act_apple:
	fall_apple(act->x, act->y);
	break;

    case act_bomb:
	blow_bomb(act->x, act->y, 0);
	break;

    case act_bonus_on:
	bonus_on();
	break;

    case act_bonus_off:
	bonus_off();
	break;

    case act_smash:
	blow_bomb(gp.x, gp.y, 1);
	break;

    case act_toolate:
	color(20, 0);
	lprint(12, 14, " OUT OF TIME! ");
	gp.Status = Status_Dead;
	break;

    case act_droplevel:
	gp.ZLevel--;
	update_jump();
	if (gp.ZLevel > 0)
	    addaction(0, 0, gp.Tid + 10.0, -1, act_droplevel);
	break;

    case act_door:
	/* Bombs can blow holes in doors, so don't assume the door is
	   still there */
	if (strchr(DOORS, screen(act->x, act->y))) {
	    color(fg(act->x, act->y), bg(act->x, act->y));
	    lprint(act->x, act->y, " ");
	    if (screen(act->x - 1, act->y) == SYM_ROCK)
		fall_rock(act->x - 1, act->y);
	}
	break;

    case act_ghost_thaw:
	{
	    int e;

	    color(28, 0);
	    for (e = 0; e < MAXGHOST; e++) {
		if (!ghost[e].dead) {
		    lprint(ghost[e].x, ghost[e].y, "\x02");
		}
	    }
	}
	break;

    case act_ghost_normal:
	gp.FS = 1;		/* Normal ghosts */
	gp.FScnt = gp.FSdiv = 1;	/* Full speed */
	break;
    }
}
