#include "compiler.h"
#include "pathname.h"
#include "grv.h"

#define APP_CONFIG "grv"
#define APP_ENVVAR "GRV_DIR"

static inline bool is_dir(const char *path)
{
    return file_type(path) == FT_Dir;
}

struct env_dir {
    const char *env;
    const char *suffix;
};

#ifdef __unix__

static const struct env_dir config_dirs[] = {
    { "XDG_CONFIG_HOME", APP_CONFIG },
    { "HOME", ".config/" APP_CONFIG },
    { "HOME", "." APP_CONFIG },
    { NULL, "." APP_CONFIG },
    { NULL, NULL }
};

#elif defined(_WIN32)

#include <libloaderapi.h>

static const struct env_dir config_dirs[] = {
    { "APPDATA", "Settings/" APP_CONFIG },
    { "APPDATA", APP_CONFIG },
    { "USERPROFILE", "Settings/" APP_CONFIG },
    { "USERPROFILE", "AppData/" APP_CONFIG },
    { "USERPROFILE", APP_CONFIG },
    { NULL, APP_CONFIG ".cfg" },
    { NULL, APP_CONFIG },
    { NULL, NULL }
};

#else

/* Last resort, hopefully straightforward */
static const struct env_dir config_dirs[] = {
    { NULL, APP_CONFIG ".cfg" },
    { NULL, APP_CONFIG },
    { NULL, NULL }
};

#endif

static char *search_env_dir(const struct env_dir *dirs)
{
    char *found  = NULL;

    for (int do_mkdir = 0; !found && do_mkdir < 2; do_mkdir++) {
	const struct env_dir *dir;
	for (dir = dirs; !found && dir->suffix; dir++) {
	    const char *env = NULL;

	    if (dir->env)
		env = getenv(dir->env);
	    if (!env)
		env = "";

	    char *path = concat_path(env, dir->suffix);

	    if (path) {
		if (is_dir(path) || (do_mkdir && !makedir(path)))
		    found = path;
		else
		    free(path);
	    }
	}
    }
    return found;
}

const char *app_config_dir(const char *set_config_dir)
{
    static char *app_dir;
    char *dir;

    /* Override configuration location, e.g. via command-line option */
    if (set_config_dir) {
	dir = strdup(set_config_dir);
	if (dir && (is_dir(dir) || !makedir(dir))) {
	    if (app_dir)
		free(app_dir);
	    return app_dir = dir;
	}
    }

    if (app_dir)
	return app_dir;

    /*
     * Always prefer $<APP_ENVVAR> if it is set in the environment,
     * even if it needs to be created.  This is done before any
     * system-specific configuration is searched.
     */
    const char *appenv = getenv(APP_ENVVAR);
    if (appenv) {
	dir = strdup(appenv);
	if (dir && (is_dir(dir) || !makedir(dir)))
	    return app_dir = dir;
    }

    dir = search_env_dir(config_dirs);
    if (!dir)
	dir = strdup("");

    return app_dir = dir;
}

const char *app_global_etcdir(void)
{
#ifdef _WIN32
    static const char *dir = NULL;

    if (dir)
	return dir;

    return dir = concat_path(getenv("ALLUSERSPROFILE"), APP_CONFIG);
#else
    return PATH_ETCDIR "/" APP_CONFIG;
#endif
}

const char *app_global_datadir(void)
{
#ifdef _WIN32
    /* On Windows, the datadir is the same as the executable location */
    static const char *dir = NULL;

    if (dir)
	return dir;

    size_t needlen = 65536;	/* Might as well start high... */
    char *exe = NULL;
    while (1) {
	exe = realloc(exe, needlen);
	if (!exe)
	    return NULL;
	size_t len = GetModuleFileName(NULL, exe, needlen);
	if (!len) {
	    free(exe);
	    return NULL;
	} else if (len < needlen) {
	    break;		/* Got it */
	}

	/* Otherwise, need a bigger buffer. Just have to keep guessing... */
	needlen <<= 1;
    }

    dir = concat_path(exe, APP_CONFIG);
    free(exe);
    return dir;
#else
    return PATH_DATADIR "/" APP_CONFIG;
#endif
}

static bool mode_is_write(const char *mode)
{
    while (*mode) {
	if (*mode == 'w')
	    return true;
	else if (*mode == ';')
	    break;
	mode++;
    }
    return false;
}

char *app_config_file(const char *file, bool for_write)
{
    char *path = concat_path(app_config_dir(NULL), file);

    /* For writing, only the user-local directory applies */
    if (for_write || file_type(path) == FT_File)
	return path;

    /* Site-local configuration */
    char *gpath = concat_path(app_global_etcdir(), file);
    if (file_type(gpath) == FT_File) {
	free(path);
	return gpath;
    }

    /* Application default configuration */
    gpath = concat_path(app_global_datadir(), file);
    if (file_type(gpath) == FT_File) {
	free(path);
	return gpath;
    }

    /* Nothing found elsewhere, stick to the apparently best choice */
    return path;
}

FILE *fopen_config(const char *file, const char *mode)
{
    char *path = app_config_file(file, mode_is_write(mode));
    if (!path) {
	errno = ENOENT;
	return NULL;
    }
    FILE *f = fopen(path, mode);
    free(path);
    return f;
}
