/*
 * graphics.c
 *
 * Graphics function emulation
 * This pretty much attempts to emulate the QuickBasic-style text screen
 * in 25x40 or 25x80 mode with no scrolling...
 */

#include "compiler.h"
#include "grv.h"
#include "graphics.h"

#define FONT_XSIZE 8
#define FONT_YSIZE 14

/* Render area size in pixels */
#define PX_WIDTH  (TS_WIDTH*FONT_XSIZE)
#define PX_HEIGHT (TS_HEIGHT*FONT_YSIZE)

/* Border area size in pixels */
#define BORDER_WIDTH  (2*FONT_XSIZE)
#define BORDER_HEIGHT FONT_YSIZE

/* Size including border area */
#define SCRN_WIDTH	(PX_WIDTH+2*BORDER_WIDTH)
#define SCRN_HEIGHT	(PX_HEIGHT+2*BORDER_HEIGHT)

typedef uint8_t font_t;
extern const font_t grv_font[256][FONT_YSIZE];

#define NCOLORS 16

/* Colors in ARGB format */
static const uint32_t colors[NCOLORS] = {
    0xff000000,			/* black */
    0xff0000aa,			/* dark blue */
    0xff00aa00,			/* dark green */
    0xff00aaaa,			/* dark cyan */
    0xffaa0000,			/* dark red */
    0xffaa00aa,			/* dark purple */
    0xffaa5500,			/* brown */
    0xffaaaaaa,			/* light grey */
    0xff555555,			/* dark grey */
    0xff5555ff,			/* light blue */
    0xff55ff55,			/* light green */
    0xff55ffff,			/* light cyan */
    0xffff5555,			/* light red */
    0xffff55ff,			/* light purple */
    0xffffff55,			/* yellow */
    0xffffffff,			/* white */
};

struct current_screen current_screen;
/* For brevity and backwards code compatibility... */
#define screendata (current_screen.data)
#define current current_screen	/* For brevity */

struct draw_surface {
    char *bufstart;		/* upper left corner of render surface */
    size_t pitch;
};
static struct draw_surface vm;

static struct main_window {
    uint32_t windowflags;
    int scale;		/* Window scale factor */
    SDL_Window *window;
    SDL_Renderer *renderer;
    SDL_Texture *texture;
    SDL_Surface *surface;
} mw;

static inline uint32_t *pixel_row(const struct draw_surface *ds,
				  unsigned int ty, unsigned int tx)
{
    return (uint32_t *)(ds->bufstart + ty * FONT_YSIZE * ds->pitch) +
	(tx * FONT_XSIZE);
}

static void render_char(int tx, int ty, const struct current_screen *scrn,
			const struct draw_surface *surf);


/*
 * Prepare screen for modification (if needed)
 * These are defined such that nested calling SHOULD be cheap.
 * (Note: probably not necessary anymore with SDL2, and thus
 * disabled entirely.)
 */

#define LOCK_SCREEN 0

static int screen_lock_ctr = 0;

static inline void lock_screen(void)
{
    if (LOCK_SCREEN && !screen_lock_ctr++)
	SDL_LockSurface(mw.surface);
}

static inline void unlock_screen(void)
{
    if (LOCK_SCREEN && !--screen_lock_ctr)
	SDL_UnlockSurface(mw.surface);
}

#ifdef HAVE_WEBP_ENCODE_H
#include <webp/encode.h>

static int screenshot_thread(void *screen_copy)
{
    struct current_screen * const scrn = screen_copy;
    time_t t = time(NULL);

    void *pixelbuf = malloc(SCRN_WIDTH*SCRN_HEIGHT*sizeof(uint32_t));
    if (!pixelbuf) {
	free(screen_copy);
	return 1;
    }

    uint32_t *p = pixelbuf;
    uint32_t *ep = p + SCRN_WIDTH*SCRN_HEIGHT;
    uint32_t border = colors[scrn->border];
    while (p < ep)
	*p++ = border;

    struct draw_surface ds;
    /* The upper left hand corner of the actual board */
    ds.bufstart = (char *)
	((uint32_t *)pixelbuf + SCRN_WIDTH*BORDER_HEIGHT + BORDER_WIDTH);
    ds.pitch = SCRN_WIDTH * sizeof(uint32_t);

    /* Render the screen into the pixelbuf, with flashing on */
    scrn->blink_mask = 0x80;

    int xs = 1 << scrn->dwidth;
    for (int y = 0; y < TS_HEIGHT; y++)
	for (int x = 0; x < TS_WIDTH; x += xs)
	    render_char(x, y, scrn, &ds);

    uint8_t *zdata = NULL;
    size_t zsize;
    zsize = WebPEncodeLosslessBGRA(pixelbuf, SCRN_WIDTH, SCRN_HEIGHT,
				   ds.pitch, &zdata);

    if (zsize && zdata) {
	struct tm tm;
	char filename[64];
	memset(&tm, 0, sizeof tm);
	localtime_r(&t, &tm);
	strftime(filename, sizeof filename, "grv-%Y%m%d-%H%M%S.webp", &tm);

	FILE *f = fopen(filename, "wb");
	if (f) {
	    fwrite(zdata, 1, zsize, f);
	    fclose(f);
	}
    }
    if (zdata)
	WebPFree(zdata);
    free(pixelbuf);
    free(screen_copy);

    return 0;
}

static void do_screenshot(void)
{
    struct current_screen *screen_copy = save_screen();
    if (!screen_copy)
	return;

    /* Everything else can be done in a subthread to not disrupt gameplay */
    SDL_CreateThread(screenshot_thread, "screenshot", screen_copy);
}

#else /* No WebP */

static void do_screenshot(void)
{
    /* Politely do nothing */
}

#endif

static bool screenshot_requested;

void request_screenshot(void)
{
    screenshot_requested = true;
}

/*
 * Update all locations on the screen that has been modified.
 * This must be called with the screen unlocked.
 *
 * This will also record a screen shot if requested.
 */
static const SDL_Rect dstrect = {
    .x = BORDER_WIDTH,
    .y = BORDER_HEIGHT,
    .w = PX_WIDTH,
    .h = PX_HEIGHT
};
static const SDL_Rect borderrects[] = {
    {
	.x = 0,
	.y = 0,
	.w = SCRN_WIDTH,
	.h = BORDER_HEIGHT
    }, {
	.x = 0,
	.y = BORDER_HEIGHT,
	.w = BORDER_WIDTH,
	.h = PX_HEIGHT
    }, {
	.x = SCRN_WIDTH - BORDER_WIDTH,
	.y = BORDER_HEIGHT,
	.w = BORDER_WIDTH,
	.h = PX_HEIGHT
    }, {
	.x = 0,
	.y = SCRN_HEIGHT - BORDER_HEIGHT,
	.w = SCRN_WIDTH,
	.h = BORDER_HEIGHT
    }
};

void update_screen(void)
{
    const uint32_t border = colors[current.border];

    /* This is inefficient, fix at some point... */
    SDL_UpdateTexture(mw.texture, NULL, mw.surface->pixels, mw.surface->pitch);
    SDL_RenderClear(mw.renderer);
    SDL_SetRenderDrawColor(mw.renderer, border >> 16, border >> 8,
			   border, 0xff);
    SDL_RenderFillRects(mw.renderer, borderrects, ARRAY_SIZE(borderrects));
    SDL_RenderCopy(mw.renderer, mw.texture, NULL, &dstrect);
    SDL_RenderPresent(mw.renderer);

    if (screenshot_requested) {
	screenshot_requested = false;
	do_screenshot();
    }
}

/*
 * Update the on-screen structure to match the screendata[]
 * for character (tx,ty), but don't refresh the rectangle just
 * yet.  This should be called with the screen locked.
 *
 * tx,ty is zero-based and must be adjusted for doublewidth mode
 */
static void put_screen(int tx, int ty)
{
    render_char(tx, ty, &current_screen, &vm);
}

static void render_char(int tx, int ty, const struct current_screen *scrn,
			const struct draw_surface *surf)
{
    const font_t *fontp;
    font_t v;
    uint32_t *pixelp, *pixelr, fgp, bgp;
    int x, y, xp, yp;
    int attr;
    const int pxw = 1 << scrn->dwidth;
    const int pxh = 1;
    const unsigned int bmask = scrn->blink_mask;

    attr = scrn->data[ty][tx].attr;
    fontp = grv_font[scrn->data[ty][tx].sym];

    bgp = colors[(attr >> 4) & 7];
    fgp = (bmask & attr) ? bgp : colors[attr & 15];

    /*
     * This assumes that either unaligned references are okay, or
     * pitch will be a multiple of 4.
     */
    pixelr = pixel_row(surf, ty, tx);

    for (y = 0; y < FONT_YSIZE; y++) {
	for (yp = 0; yp < pxh; yp++) {
	    pixelp = pixelr;
	    v = fontp[y];
	    for (x = 0; x < FONT_XSIZE; x++) {
		uint32_t pixel;
		pixel = v & ((font_t) 1 << (FONT_XSIZE - 1)) ? fgp : bgp;
		v <<= 1;
		for (xp = 0; xp < pxw; xp++)
		    *pixelp++ = pixel;
	    }
	    pixelr = (uint32_t *) ((char *)pixelr + surf->pitch);
	}
    }
}

/*
 * This routine switches the blink status, then goes around the screen
 * and updates all characters which has the blink attribute set.
 */
void update_blink(void)
{
    int x, y, xs;

    xs = current.dwidth + 1;

    current.blink_mask ^= 0x80;

    lock_screen();
    for (y = 0; y < TS_HEIGHT; y++)
	for (x = 0; x < TS_WIDTH; x += xs)
	    if (screendata[y][x].attr & 0x80)
		put_screen(x, y);
    unlock_screen();

    update_screen();
}

static void gputc(unsigned char ch)
{
    const uint8_t attr = current.attr;
    const int xwid = current.dwidth + 1;
    int x = current.x;
    int y = current.y;

    lock_screen();

    screendata[y][x].sym = ch;
    screendata[y][x].attr = attr;
    put_screen(x, y);
    x += xwid;
    if (x >= TS_WIDTH) {
	x = 0;
	y++;
	if (y >= TS_HEIGHT)
	    y = 0;
    }

    current.x = x;
    current.y = y;

    unlock_screen();
}

/*
 * Display a string of characters and advance cursor
 */
void print(const char *str)
{
    const unsigned char *p = (const unsigned char *)str;
    unsigned char ch;

    if (!str || !*str)
	return;

    lock_screen();

    while ((ch = *p++))
	gputc(ch);

    unlock_screen();
}

/*
 * Same except with printf-style formatting
 */
void gprintf(const char *fmt, ...)
{
    char buffer[TS_WIDTH*TS_HEIGHT+1];
    va_list ap;

    va_start(ap, fmt);
    vsnprintf(buffer, sizeof buffer, fmt, ap);
    buffer[sizeof buffer-1] = 0; /* Officially unnecessary but paranoia... */
    va_end(ap);
    print(buffer);
}

/*
 * locate() and print() as a combined function
 */
void lprint(int r, int c, const char *str)
{
    locate(r, c);
    print(str);
}

/*
 * locate() and print a single character
 */
void lputc(int r, int c, int ch)
{
     locate(r, c);
     gputc(ch);
}

/*
 * locate() and print a single character any number of timers
 */
void lputcn(int r, int c, int ch, unsigned int n)
{
     locate(r, c);
     lock_screen();
     while (n--)
	 gputc(ch);
     unlock_screen();
}

/*
 * locate() and gprintf() as a combined function
 */
void lprintf(int r, int c, const char *fmt, ...)
{
    char buffer[TS_WIDTH*TS_HEIGHT+1];
    va_list ap;

    locate(r, c);
    va_start(ap, fmt);
    vsnprintf(buffer, sizeof buffer, fmt, ap);
    buffer[sizeof buffer-1] = 0; /* Officially unnecessary but paranoia... */
    va_end(ap);
    print(buffer);
}

/*
 * Set 40/80 character mode.  Unlike hardware this doesn't change the
 * text currently on the screen, but funny things can happen; cls() is
 * highly recommended.
 */
void width(int wid)
{
    switch (wid) {
    case 40:
	current.dwidth = 1;
	current.x &= ~1;
	break;
    case 80:
	current.dwidth = 0;
	break;
    default:
	abort();
    }
}

/*
 * Clear the screen, set all characters to ' ' and set all pixels to the
 * current background, then refresh the screen content including the
 * border.
 */
void cls(void)
{
    int i;
    struct screen_char *scp = &screendata[0][0];
    struct screen_char empty;
    uint32_t pixelval;
    SDL_Rect rect;

    lock_screen();

    if (opt.debug)
	fprintf(stderr, "cls with attr %02x\n", current.attr);

    empty.sym = ' ';
    empty.attr = current.attr;

    current.border = (current.attr >> 4) & 7; /* Set border to new background */
    pixelval = colors[current.border];

    for (i = TS_WIDTH * TS_HEIGHT; i; i--)
	*scp++ = empty;

    rect.x = rect.y = 0;
    rect.w = PX_WIDTH;
    rect.h = PX_HEIGHT;

    SDL_FillRect(mw.surface, &rect, pixelval);

    unlock_screen();

    if (opt.debug) {
	fprintf(stderr, "x = %d, y = %d, w = %d, h = %d\n",
		rect.x, rect.y, rect.w, rect.h);
    }

    update_screen();

    current.x = current.y = 0;
}

/*
 * Save the screen data in a malloc'd buffer
 */
struct current_screen *save_screen(void)
{
    struct current_screen *save;

    save = malloc(sizeof *save);
    if (!save)
	return NULL;

    return memcpy(save, &current_screen, sizeof *save);
}

void restore_screen(const struct current_screen *save)
{
    int x, y, xs;

    memcpy(&current_screen, save, sizeof current_screen);

    xs = current.dwidth + 1;

    lock_screen();

    for (y = 0; y < TS_HEIGHT; y++) {
	for (x = 0; x < TS_WIDTH; x += xs) {
	    put_screen(x, y);
	}
    }

    unlock_screen();
    update_screen();
}

static const Uint32 fullscreen = SDL_WINDOW_FULLSCREEN_DESKTOP;

void screen_togglefullscreen(void)
{
    mw.windowflags ^= fullscreen;
    SDL_SetWindowFullscreen(mw.window, mw.windowflags & fullscreen);
}

static int max_window_scale(void)
{
    int display = SDL_GetWindowDisplayIndex(mw.window);
    SDL_Rect rect;

    if (SDL_GetDisplayBounds(display, &rect))
	return -1;

    return min(rect.w/SCRN_WIDTH, rect.h/SCRN_HEIGHT);
}

void screen_zoom(int change)
{
    int scale = mw.scale + change;

    if (scale <= 1) {
	scale = 1;
    } else {
	int maxscale = max_window_scale();

	if (scale > maxscale)
	    scale = maxscale;
    }

    /* change == 0 means force setting the window size */
    if (scale == mw.scale && change)
	return;

    mw.scale = scale;

    int xpix = SCRN_WIDTH  * scale;
    int ypix = SCRN_HEIGHT * scale;

    SDL_SetWindowSize(mw.window, xpix, ypix);
}

/*
 * Compute scaling factors and frame
 */
static int configure_screen(int windowed)
{
    int xpix, ypix;

    mw.windowflags = SDL_WINDOW_ALLOW_HIGHDPI;
    mw.scale = windowed ? windowed : 1;
    xpix = SCRN_WIDTH  * mw.scale;
    ypix = SCRN_HEIGHT * mw.scale;
    if (!windowed) {
	/* full screen */
	mw.windowflags |= fullscreen;
    }

    mw.window = SDL_CreateWindow("Grävning!",
				 SDL_WINDOWPOS_CENTERED,
				 SDL_WINDOWPOS_CENTERED,
				 xpix, ypix, mw.windowflags);
    if (!mw.window)
	return -1;

    /* No mouse cursor, please */
    SDL_ShowCursor(opt.debug ? SDL_ENABLE : SDL_DISABLE);

    mw.renderer = SDL_CreateRenderer(mw.window, -1, SDL_RENDERER_PRESENTVSYNC);
    if (!mw.renderer)
	return -1;

    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");
    SDL_RenderSetLogicalSize(mw.renderer, SCRN_WIDTH, SCRN_HEIGHT);

    mw.texture = SDL_CreateTexture(mw.renderer,
				   SDL_PIXELFORMAT_ARGB8888,
				   SDL_TEXTUREACCESS_STREAMING,
				   PX_WIDTH, PX_HEIGHT);
    if (!mw.texture)
	return -1;

    mw.surface = SDL_CreateRGBSurface(0, PX_WIDTH, PX_HEIGHT, 32,
				      0xff << 16, 0xff << 8, 0xff, 0xff << 24);
    if (!mw.surface)
	return -1;

    vm.bufstart = mw.surface->pixels;
    vm.pitch = mw.surface->pitch;

    cls();
    return 0;
}

/*
 * Initialize SDL and the data structures
 */
int screen_init(int windowed)
{
    if (SDL_Init(SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO |
		 /* SDL_INIT_JOYSTICK|SDL_INIT_HAPTIC|SDL_GAMECONTROLLER| */
		 SDL_INIT_EVENTS))
	return -1;

    atexit(SDL_Quit);

    if (configure_screen(windowed))
	return -1;

    timer_periodic(400, event_blink);

    return 0;
}
