/*
 * graphics.h
 *
 * Graphics function emulation
 * This pretty much attempts to emulate the QuickBasic-style text screen
 * in 25x40 or 25x80 mode with no scrolling...
 */

#ifndef GRAPHICS_H
#define GRAPHICS_H 1

#include "compiler.h"

int screen_init(int window);
void print(const char *str);
void lputc(int r, int c, int ch);
void lputcn(int r, int c, int ch, unsigned int count);
void lprint(int r, int c, const char *str);
void gprintf(const char *fmt, ...);
void lprintf(int r, int c, const char *fmt, ...);
void width(int wid);
void cls(void);
void update_blink(void);
void update_screen(void);
struct current_screen;
struct current_screen *save_screen(void);
void restore_screen(const struct current_screen *);

enum setstate {
    set_off    = 0,
    set_on     = 1,
    set_toggle = 2
};
void screen_togglefullscreen(void);
void screen_zoom(int);

void request_screenshot(void);

/*
 * Reading from the screen buffer is very common, so make these inlines.
 */

#define TS_WIDTH  80
#define TS_HEIGHT 25

struct screen_char {
    uint8_t sym;		/* Symbol code */
    uint8_t attr;		/* Attribute code */
};
struct current_screen {
    int x, y;			/* Location of the output "cursor" */
    uint8_t attr;		/* Current output attribute */
    uint8_t blink_mask;		/* Current blink enable/disable */
    uint8_t dwidth;		/* Double width characters, must be 0 or 1 */
    uint8_t border;		/* Current border color */
    struct screen_char data[TS_HEIGHT][TS_WIDTH];
};
extern struct current_screen current_screen;

/*
 * Set the attribute: fg, bg
 * For compatiblity's sake the flashing attribute is in fg, not bg
 */
static inline void color(int fg, int bg)
{
    current_screen.attr = (fg & 15) | ((bg & 7) << 4) | ((fg & 16) << 3);
}

/*
 * Set the cursor position (1-based for compatiblity)
 */
static inline void locate(int r, int c)
{
    current_screen.y = r - 1;
    current_screen.x = (c - 1) << current_screen.dwidth;
}

/*
 * Get the screen_char structure at a specific coordinate; returns
 * a dummy structure if out of range.
 */
static inline struct screen_char screenx(int r, int c)
{
    static const struct screen_char dummy = { 0, 0 };
    unsigned int rr = r - 1;
    unsigned int cc = (c - 1) << current_screen.dwidth;
    if (rr >= TS_HEIGHT || cc >= TS_WIDTH)
	return dummy;

    return current_screen.data[rr][cc];
}

/*
 * Get the symbol at a specific screen coordinate
 */
static inline int screen(int r, int c)
{
    return screenx(r,c).sym;
}

/*
 * Get the attribute at a specific screen coordinate
 */
static inline int screen1(int r, int c)
{
    return screenx(r,c).attr;
}

/* Foreground color of a specific character, sans blink */
static inline int fg(int r, int c)
{
    return screen1(r, c) & 15;
}

/* Background color of a specific character */
static inline int bg(int r, int c)
{
    return (screen1(r, c) >> 4) & 7;
}

/* Blink status of a specific character */
static inline bool flashing(int r, int c)
{
    return (int8_t)screen1(r, c) < 0;
}

#endif				/* GRAPHICS_H */
