/*
 * loadmap.c
 *
 * Load user file data, and draw a level from it
 */
#include "compiler.h"
#include "graphics.h"
#include "grv.h"
#include "grvfiledata.h"
#include "highscore.h"
#include "pathname.h"

#include <zlib.h>
#include <ctype.h>

int ELev;
static const struct grvfileheader **filedata;

/*
 * Configure the maximum level count and allocate the filedata array
 */
void set_levels(int count)
{
    const struct grvfileheader *last = ELev >= 2 ? filedata[ELev - 2] : NULL;
    int old_ELev = ELev;

    /* The user is always right, long live the user */
    if (opt.endlevel)
	count = opt.endlevel;

    /* Well, almost... */
    if (count < 2)
	count = 2;
    else if (count > 999)
	count = 999;

    ELev = count;

    if (ELev != MAX_LEVEL)
	opt.cheat = true;	/* No high scores for customizations yet */

    filedata = realloc(filedata, ELev * sizeof(*filedata));

    for (int i = old_ELev; i < ELev - 1; i++)
	filedata[i] = last;

    filedata[ELev - 1] = NULL;	/* The last level is always builtin */
}

#define TimerBits 6
#define MaxTimers (1 << TimerBits)
#define TimerMask ((1 << TimerBits)-1)

/*
 * Load map data into memory from a compressed data buffer, passing
 * in a pointer to struct grvzfileheader which need not be contiguous.
 *
 * If baselevel (1-based!) is nonzero, use it instead of the base level
 * specified in the file data.
 *
 * Returns false on error.
 */
static inline const void *refptr(const struct grvfileheader *hdr, size_t offset)
{
    return (const void *)((const char *)hdr + offset);
}

bool load_map_data(const struct grvzfileheader *zhdr, const void *zdata,
		   int baselevel, int endlevel)
{
    void *data;
    z_stream zs;
    uint32_t zsize = SDL_SwapBE32(zhdr->zsize);
    uint32_t dsize = SDL_SwapBE32(zhdr->dsize);

    if (baselevel >= ELev)
	return false;

    if (!endlevel)
	endlevel = ELev - 1;

    if (endlevel < baselevel)
	return false;

    if (memcmp(zhdr->magic, GRVFILE_MAGIC, sizeof zhdr->magic))
	return false;

    if (dsize < sizeof(struct grvfileheader))
	return false;

    if (zsize > GRV_MAX_DATA || dsize > GRV_MAX_DATA)
	return false;		/* Safety check */

    data = malloc(dsize);
    if (!data)
	return false;

    memset(&zs, 0, sizeof zs);
    zs.avail_in = zsize;
    zs.next_in = (void *)zdata;
    zs.avail_out = dsize;
    zs.next_out = data;

    if (inflateInit(&zs) < 0)
	return false;

    if (inflate(&zs, Z_FINISH) != Z_STREAM_END)
	return false;

    if (zs.total_out != dsize)
	return false;

    const struct grvfileheader *hdr = data;

    if (memcmp(hdr->magic, GRVDATA_MAGIC, sizeof hdr->magic) ||
	hdr->major != GRVDATA_MAJOR ||
	hdr->incompatflags ||
	hdr->headersize < sizeof *hdr ||
	hdr->boardoffset < sizeof *hdr ||
	hdr->randoffset < sizeof *hdr ||
	hdr->xsize != 40 || hdr->ysize != 22 ||
	hdr->timerbits != TimerBits ||
	hdr->usedtimers > MaxTimers ||
	hdr->usedtimers > hdr->randobjs || hdr->bombtimer > hdr->usedtimers)
	return false;

    if (baselevel < 1)
	baselevel = hdr->baselevel;

    baselevel--;
    endlevel--;
    if (baselevel < 0)
	return false;
    endlevel = min(endlevel, ELev - 2);
    if (endlevel < baselevel)
	return false;

    /*
     * This data will be used for the base level upwards
     * In the future, consider loading first, then sorting by
     * base level.
     */
    for (int l = baselevel; l <= endlevel; l++)
	filedata[l] = hdr;

    return true;
}

bool draw_data_level(void)
{
    const struct grvfileheader *hdr = filedata[gp.Level];
    const struct grvfileitem *item;
    const struct grvfilerandom *rnd, *tmr;

    if (!hdr)
	return false;		/* Draw default level */

    struct grvfilerandom *timers;
    timers = calloc(MaxTimers, sizeof *timers);

    /* Extract timers from the random data */
    rnd = refptr(hdr, hdr->randoffset);
    for (unsigned int i = 0; i < hdr->randobjs; i++, rnd++) {
	if ((unsigned int)(rnd->minlvl - 1) > (unsigned int)gp.Level)
	    continue;		/* Disabled whatever it is */

	if (rnd->rndflags & GFR_TIMER) {
	    struct grvfilerandom *tp;
	    if (rnd->itemtype >= hdr->usedtimers)
		continue;	/* Shouldn't happen */

	    tp = &timers[rnd->itemtype];

	    tp->itemtype = rnd->itemtype;
	    tp->rndflags = rnd->rndflags;
	    tp->minlvl = rnd->minlvl;
	    tp->valmax = rnd->valmax;

	    if (rnd->rndflags & GFR_MULTI) {
		tp->valbase = rnd->valbase;
		tp->vallvl = rnd->vallvl;
		tp->valrnd = rnd->valrnd;
	    } else {
		/* Force the value to be resolved once */
		tp->valbase = calc_random(rnd);
		tp->vallvl = 0.0;
		tp->valrnd = 0.0;
	    }
	}
    }

    /* High scores for custom levels not supported yet */
    gp.Cheat = 1;

    /* Draw the fixed part of the board */
    item = refptr(hdr, hdr->boardoffset);
    for (int r = 0; r < hdr->ysize; r++) {
	for (int c = 0; c < hdr->xsize; c++) {
	    if (item->timerprio >> TimerBits)
		tmr = &timers[item->timerprio & TimerMask];
	    else
		tmr = NULL;
	    drawfixeditem(r, c, item++, tmr);
	}
    }

    /* Draw random items */
    rnd = refptr(hdr, hdr->randoffset);
    tmr = &timers[hdr->bombtimer];	/* Only one that matters */

    for (unsigned int i = 0; i < hdr->randobjs; i++, rnd++) {
	if (!(rnd->rndflags & GFR_TIMER))
	    drawrandomitems(rnd, tmr);
    }

    free(timers);

    /* Hey, it worked! */
    return true;
}

static void *read_file(const char *filename, size_t *size_p)
{
    off_t len;
    FILE *f = NULL;
    void *data = NULL;

    *size_p = 0;

    f = fopen(filename, "rb");
    if (!f)
	goto err;

    if (fseeko(f, 0, SEEK_END))
	goto err;

    len = ftello(f);
    if (len == (off_t) - 1)
	goto err;

    if (len > GRV_MAX_DATA)
	goto err;

    if (fseeko(f, 0, SEEK_SET))
	goto err;

    *size_p = len;

    data = malloc(len + 1);
    if (!data)
	goto err;

    if (fread(data, 1, len, f) != (size_t)len)
	goto err;

    ((char *)data)[len] = 0;	/* Zero-terminate the buffer */

done:
    if (f)
	fclose(f);

    return data;

err:
    if (data) {
	free(data);
	data = NULL;
    }
    goto done;
}

int load_map_file(const char *filename, int baselevel, int endlevel)
{
    char *fn = strdup(filename ? filename : "");
    void *data = NULL;
    int err;

    if (!fn)
	goto err;

    errno = EINVAL;

    if (baselevel < 0) {
	/* Look for @<level> suffix */
	char *ep = strchr(fn, '\0') - 1;
	while (ep > fn && isdigit(*ep))
	    ep--;
	if (*ep == '@') {
	    baselevel = atoi(ep + 1);
	    *ep = '\0';
	} else {
	    baselevel = 0;
	}
    }

    if (baselevel > ELev - 1)
	goto err;

    if (endlevel < 1 || endlevel > ELev - 1)
	endlevel = ELev - 1;

    if (!fn[0] || (fn[0] == '*' && !fn[1])) {
	/* Clear this range of levels, resetting them to builtin */
	if (baselevel < 1)
	    baselevel = 1;

	memset(&filedata[baselevel - 1], 0,
	       (endlevel - baselevel + 1) * sizeof(*filedata));
	errno = 0;
	goto err;
    }

    size_t size;
    data = read_file(fn, &size);
    if (!data)
	goto err;

    const struct grvzfileheader *zhdr = data;
    const void *zdata = &zhdr[1];

    if (size < sizeof *zhdr || size < SDL_SwapBE32(zhdr->zsize))
	goto err;

    if (!custom.title)
	custom.title = "With custom maps";

    if (load_map_data(zhdr, zdata, baselevel, endlevel))
	errno = 0;
err:
    err = errno;
    if (data)
	free(data);
    if (fn)
	free(fn);

    return err;
}

static inline bool is_eol(int c)
{
    return c == '\n' || c == '\r';
}

static inline bool is_blank(int c)
{
    return !is_eol(c) && isspace(c);
}

static char *skipspace(char *ptr)
{
    if (!ptr)
	return NULL;
    while (1) {
	if (!*ptr)
	    return NULL;
	if (!is_blank(*ptr))
	    return ptr;
	ptr++;
    }
}

static char *looking_at(const char *str, char *ptr)
{
    if (!ptr)
	return NULL;

    while (*str) {
	if (*str++ != *ptr++)
	    return NULL;
    }
    if (!isspace(*ptr))
	return NULL;

    return skipspace(ptr);
}

static char *next_line(char *ptr)
{
    if (!ptr)
	return NULL;
    while (1) {
	if (!*ptr)
	    return NULL;
	if (is_eol(*ptr))
	    return ptr + 1;
	ptr++;
    }
}

static char *get_filename(char **pp)
{
    char *start, *p;
    p = start = *pp;

    if (!p || !*p || is_eol(*p))
	return NULL;

    if (*p == '\'' || *p == '\"') {
	char delim = *p++;
	start = p;
	while (*p && *p != delim)
	    p++;
    } else {
	while (*p && !is_eol(*p) && !is_blank(*p))
	    p++;
    }

    /* Now p should point to the first character beyond the filename */
    *pp = p;
    return start;
}

/* Load a customization string for display (limited to 40 characters) */
static void custom_string(char *e, char **pp, const char **dest)
{
    *pp = e;
    e = get_filename(pp);
    if (e)
	*dest = strndup(e, min(*pp-e,40));
}

/*
 * Game scripts are searched for as follows:
 * - in the current directory, by explicit name
 * - in the config directory
 *
 * In each case, try to append .grv at the end as part of the
 * search.
 *
 * If a directory is encountered, try to append the basename, e.g.
 * <game> -> <game>, <game>.grv, <game>/<game>, <game>/<game>.grv
 *
 * Returns the resulting file path in a malloc()'d buffer, or NULL
 * if no file was found.
 */
static char *file_search(const char *name)
{
    const char * const dirs[] = {
	"", app_config_dir(NULL), app_global_etcdir(), app_global_datadir()
    };
    const char * const subdirs[] = {
	"", "games/"
    };
    static const char * const suff[] = { "", ".grv" };
    enum file_type type;
    char *path = NULL;

    /* If the filename is absolute, don't search other directories */
    size_t maxdir = filename_is_absolute(name) ? 1 : ARRAY_SIZE(dirs);

    for (size_t i = 0; !path && i < maxdir; i++) {
	char *bpath = concat_path(dirs[i], name);

	for (size_t j = 0; j < ARRAY_SIZE(subdirs); j++) {
	    char *tpath = concat_path(bpath, subdirs[j]);

	    for (size_t k = 0; k < ARRAY_SIZE(suff); k++) {
		char *epath;

		asprintf(&epath, "%s%s", tpath, suff[j]);
		if (!epath)
		    break;

		type = file_type(epath);
		if (type == FT_File) {
		    path = epath;
		    break;
		}

		/* Try to see if it is a subdirectory */
		char *opath = epath;
		const char *obase = basename_ptr(opath);
		epath = concat_path(tpath, obase);
		free(opath);
		if (file_type(epath) == FT_File) {
		    path = epath;
		    break;
		}

		free(epath);
	    }
	    free(tpath);
	}
	free(bpath);
    }

    return path;
}

int load_game_script(const char *filename)
{
    size_t size;
    char *txt = NULL;
    int e;
    char *path = file_search(filename);

    if (!path)
	goto err;

    errno = 0;
    txt = read_file(path, &size);
    if (!txt)
	goto err;

    if (!custom.title)
	custom.title = "Custom game";

    char *p = txt;

    do {
	char *e;
	p = skipspace(p);
	if (!p || is_eol(*p) || *p == '#') {
	    continue;
	} else if ((e = looking_at("map", p))) {
	    int from, to;
	    if (*e == '-' || isdigit(*e)) {
		if (*e == '-')
		    from = 1;
		else
		    from = strtoul(e, &e, 10);
		if (*e == '-') {
		    e++;
		    if (isdigit(*e))
			to = strtoul(e, &e, 10);
		    else
			to = ELev - 1;
		} else {
		    to = from;
		}
		e = skipspace(e);
		if (!e)
		    break;
	    } else {
		from = to = -1;	/* Use the header information from the file */
	    }

	    p = e;
	    const char *file = get_filename(&p);
	    if (!file) {
		load_map_file(NULL, from, to);
	    } else {
		char ec = *p;
		*p = '\0';
		char *fpath = concat_path(path, file);
		*p = ec;
		if (fpath) {
		    load_map_file(fpath, from, to);
		    free(fpath);
		}
	    }
	} else if ((e = looking_at("levels", p)) && isdigit(*e)) {
	    set_levels(strtoul(e, NULL, 10));
	    continue;
	} else if ((e = looking_at("esc", p)) || (e = looking_at("escape", p))) {
	    EscStart = atoi(e) - 1;
	} else if ((e = looking_at("title", p))) {
	    custom_string(e, &p, &custom.title);
	} else if ((e = looking_at("byline", p))) {
	    custom_string(e, &p, &custom.byline);
	} else if ((e = looking_at("author", p))) {
	    custom_string(e, &p, &custom.author);
	}
    } while ((p = next_line(p)));

err:
    e = errno;
    if (txt)
	free(txt);
    if (path)
	free(path);
    return e;
}
