/*
 * Very simple file path manipulation functions
 */

#include "compiler.h"
#include "pathname.h"

/*
 * This is a valid path separator on all supported platforms, *IF*
 * it is preceeded with a valid directory name which does *not* end
 * in a path separator character. Otherwise this may end up doing
 * the wrong thing, like changing c: to c:/ on Windows.
 *
 * Note that is_path_separator() can recognize other characters too.
 */
#define PATH_SEPARATOR '/'

/*
 * Returns the first character after the last path separator, if any.
 * It is a basic filename if the return pointer == the passed-in pointer.
 */
char *basename_ptr(const char *path)
{
    const char *p = strchr(path, '\0');
    while (p > path && !is_path_separator(p[-1]))
	p--;
    return (char *)p;
}

/*
 * Path string for "this directory" - valid on all supported platforms.
 */
#define THIS_DIR "."
#define THIS_DIR_LEN (sizeof(THIS_DIR)-1)

enum file_type file_type(const char *path)
{
    struct stat st;

    if (!path || !*path || stat(path, &st))
	return FT_None;

    if (S_ISREG(st.st_mode))
	return FT_File;

    if (S_ISDIR(st.st_mode))
	return FT_Dir;

    return FT_Special;
}

/*
 * Concatenate two paths. If dir refers to a non-directory, try to
 * strip the file part of the name - this allows one file to be the
 * pathname base for another.
 */
char *concat_path(const char *dir, const char *file)
{
    size_t dl, fl;
    char *path = NULL;
    char *pdir = NULL;

    if (!dir)
        dir = "";
    if (!file)
	file = THIS_DIR;

    if (filename_is_absolute(file) || !*dir)
	return strdup(file);

    dl = strlen(dir);

    if (!is_path_separator(dir[dl-1]) && file_type(dir) != FT_Dir &&
	(pdir = malloc(dl + THIS_DIR_LEN + 1))) {
	char *p = pdir+dl;

	memcpy(pdir, dir, dl);

	while (p > pdir && !is_path_separator(p[-1]))
	    p--;
	memcpy(p, THIS_DIR, THIS_DIR_LEN+1);

	if (file_type(pdir) == FT_Dir) {
	    *p = '\0';
	    dir = pdir;
	    dl = p - pdir;
	}
    }

    fl = strlen(file);

    path = malloc(dl + fl + 2);	/* One byte for /, one for \0 */
    if (path) {
	char *p = path;

	memcpy(p, dir, dl);
	p += dl;
	if (!is_path_separator(p[-1]))
	    *p++ = PATH_SEPARATOR;
	memcpy(p, file, fl+1);
    }

    if (pdir)
	free(pdir);

    return path;
}
