/*
 * prefs.c
 *
 * Select preferences
 */

#include "graphics.h"
#include "grv.h"

/*
 * Keyboard map
 */

const union kbd_keys default_kbd_keys = {
    .move.up = SDLK_w,		/* Up */
    .move.left = SDLK_a,	/* Left */
    .move.right = SDLK_s,	/* Right */
    .move.down = SDLK_z,	/* Down */
    .shoot.up = SDLK_u,		/* Shoot up */
    .shoot.left = SDLK_h,	/* Shoot left */
    .shoot.right = SDLK_j,	/* Shoot right */
    .shoot.down = SDLK_n,	/* Shoot down */
    .stop = SDLK_SPACE,		/* Stop */
    .esc = SDLK_ESCAPE,		/* Escape */
    .pause = SDLK_F1,		/* Pause */
    .fullscreen = SDLK_F11,	/* Toggle full screen */
    .screenshot = SDLK_F2,	/* Capture screenshot */
};

union kbd_keys kbd_keys;

/* Collapse and normalize whitespace; returns true if \n terminator found */
static bool cleanspace(char *p)
{
    char *q = p;
    bool newline = false;

    while (*p && *p != '\n' && isspace(*p))
	p++;
    while (*p) {
	if (*p == '\n') {
	    newline = true;
	    break;
	} else if (isspace(*p)) {
	    if (*q != ' ')
		*q++ = ' ';
	} else {
	    *q++ = *p;
	}
	p++;
    }
    *q = '\0';
    return newline;
}

static void load_keys(void)
{
    FILE *f;
    char linebuf[512];

    f = fopen_config("grvkeys.dat", "rt");
    if (!f)
	return;

    unsigned int nkey = 0;
    while (fgets(linebuf, sizeof linebuf, f)) {
	SDL_Keycode key;
	int pos;

	if (!cleanspace(linebuf))
	    break;

	key = SDLK_UNKNOWN;
	if (sscanf(linebuf, "key %u %n", &nkey, &pos) == 1) {
	    /* New interface - use SDL key names */
	    key = SDL_GetKeyFromName(linebuf + pos);
	} else {
	    char *q;
	    unsigned long keycode = strtoul(linebuf, &q, 10);
	    if (q != linebuf && !*q)
		key = keycode;
	}

	if (key != SDLK_UNKNOWN && nkey < NKEYS) {
	    /* nkey++ to support the legacy format */
	    kbd_keys.keys[nkey++] = key;
	}
    }
    fclose(f);
}

static void save_keys(void)
{
    FILE *f;
    int i;

    f = fopen_config("grvkeys.dat", "wt");
    if (!f)
	return;

    for (i = 0; i < NKEYS; i++) {
	fprintf(f, "key %d %s\n", i, SDL_GetKeyName(kbd_keys.keys[i]));
    }
    fclose(f);
}

void load_prefs(void)
{
    kbd_keys = default_kbd_keys;
    load_keys();
}

static void clear_keys(union kbd_keys *keys)
{
    for (int i = 0; i < NKEYS; i++)
	keys->keys[i] = -1;	/* Impossible */
}

static void draw_keys(const union kbd_keys *keys)
{
    color(0, gp.c);
    for (int i = 0; i < NKEYS; i++)
	lprintf(i + 3, 24, " %-12s ", SDL_GetKeyName(keys->keys[i]));
}

static void do_edit_keys(union kbd_keys *keys)
{
    lputcn(20, 1, ' ', 200);	/* Clear menu */

    for (int i = 0; i < NKEYS; i++) {
	color(0, gp.c2);
	lprintf(i + 3, 24, " %-12s ", SDL_GetKeyName(keys->keys[i]));

	keys->keys[i] = get_key()->keysym.sym;
	color(0, gp.c);
	lprintf(i + 3, 24, " %-12s ", SDL_GetKeyName(keys->keys[i]));
    }
}

static bool keys_bad(const union kbd_keys *keys)
{
    for (int i = 0; i < NKEYS-1; i++)
	for (int j = i+1; j < NKEYS; j++)
	    if (keys->keys[i] == keys->keys[j])
		return true;

    return false;
}

void edit_keys(void)
{
    bool changed, edited;
    union kbd_keys user_keys = kbd_keys; /* Original user keys */
    union kbd_keys edit_keys = kbd_keys; /* Edited user keys */

    initscreen(40);

    lprint(3, 4, "Move up . . . . . .");
    lprint(4, 4, "Move left . . . . .");
    lprint(5, 4, "Move right  . . . .");
    lprint(6, 4, "Move down . . . . .");
    lprint(7, 4, "Shoot up  . . . . .");
    lprint(8, 4, "Shoot left  . . . .");
    lprint(9, 4, "Shoot right . . . .");
    lprint(10, 4, "Shoot down  . . . .");
    lprint(11, 4, "Dead stop . . . . .");
    lprint(12, 4, "Quick escape  . . .");
    lprint(13, 4, "Pause game  . . . .");
    lprint(14, 4, "Full screen . . . .");
    lprint(15, 4, "Screenshot  . . . .");

    /* Suppress all special keystrokes in the active keymap */
    clear_keys(&kbd_keys);

    changed = edited = false;
    while (1) {
	bool bad = keys_bad(&edit_keys);

	draw_keys(&edit_keys);
	if (bad) {
	    color(15, gp.c);
	    lprint(18, 5, "ERROR: Duplicated keys");
	}

	color(0, gp.c);

	lputcn(20, 1, ' ', 200);
	lprint(20, 5, "E ... Edit");
	if (edited)
	    print(" again");
	lprint(21, 5, "D ... Load defaults");
	lprint(22, 5, "Q ... Quit");
	if (changed)
	    print(" and discard");
	if (changed && !bad)
	    lprint(23, 5, "S ... Save");

	SDL_Keycode sym = 0;

	while (!sym) {
	    switch ((sym = get_key()->keysym.sym)) {
	    case SDLK_e:
		do_edit_keys(&edit_keys);
		changed = edited = true;
		break;
	    case SDLK_d:
		edit_keys = default_kbd_keys;
		changed = true;
		edited = false;
	    break;
	    case SDLK_q:
		kbd_keys = user_keys;
		return;
	    case SDLK_s:
		if (changed && !bad) {
		    kbd_keys = edit_keys;
		    save_keys();
		    return;
		}
		/* Fall through */
	    default:
		sym = 0;	/* Unknown key, no redraw */
		break;
	    }
	}
    }
}
