/*
 * utils.c
 *
 * Utility functions
 */

#include "graphics.h"
#include "grv.h"

/* Format time elapsed in MM:SS.s format */
char *format_time(double t)
{
    static char ts[32];
    int ds = (int)(t * 10.0);	/* Deciseconds */

    sprintf(ts, "%02d:%02d.%01d", ds / 600, (ds % 600) / 10, ds % 10);

    return ts;
}

/*
 * Convenience function to push a user event
 */
static void push_event(enum user_event code)
{
    SDL_Event event;
    event.type = SDL_USEREVENT;
    event.user.code = code;
    event.user.data1 = event.user.data2 = NULL;

    SDL_PushEvent(&event);
}

/* Standard callback routine from a timer to post a periodic user event */
/* (int)param is the event code */
uint32_t post_periodic(uint32_t interval, void *param)
{
    push_event((size_t)param);
    return interval;
}

/* Standard callback routine from a timer to post a one-shot user event */
/* (int)param is the event code */
uint32_t post_oneshot(uint32_t interval, void *param)
{
    push_event((size_t)param);
    (void)interval;
    return 0;
}

timer_id timer_periodic(uint32_t ms, enum user_event event)
{
    return SDL_AddTimer(ms, post_periodic, (void *)event);
}

timer_id timer_oneshot(uint32_t ms, enum user_event event)
{
    return SDL_AddTimer(ms, post_oneshot, (void *)event);
}

/* Very simple event handler which just posts a timer and handles
   blink events until the timer expires */
void mymssleep(int ms)
{
    timer_id mytimer = timer_oneshot(ms, event_sleep);
    SDL_Event event;

    update_screen();

    while (wait_event(&event)) {
	if (event.type == SDL_USEREVENT &&
	    event.user.code == event_sleep)
	    break;
    }
    timer_cancel(mytimer);
}

/*
 * Get or poll for an event, process certain kinds of events that are
 * handled the same way throughout the code.
 */
int wait_event(SDL_Event *event)
{
    SDL_Event dummy_event;
    if (!event)
	event = &dummy_event;

    while (SDL_WaitEvent(event)) {
	switch (event->type) {
	case SDL_USEREVENT:
	    switch (event->user.code) {
	    case event_blink:
		update_blink();
		break;
	    default:
		return 1;
	    }
	    break;
	case SDL_KEYDOWN:
	{
	    SDL_Keymod  mod = event->key.keysym.mod;
	    SDL_Keycode sym = event->key.keysym.sym;
	    int key_exit = 1;

	    if (mod & KMOD_ALT) {
		switch (sym) {
		case SDLK_RETURN:	/* Alt-Enter */
		    screen_togglefullscreen();
		    break;
		case SDLK_END:		/* Immediately zap the game */
		    exit(0);

		case SDLK_q:		/* Immediate death */
		    gp.Life = 0;
		    /* Fall through */
		case SDLK_w:		/* Immediate death (one life) */
		    if (gp.EOLWait)
			key_exit = 0; /* As if we had gotten a failure */
		    gp.EOLWait = 0;
		    /* Fall through */
		case SDLK_d:		/* Normal death */
		    gp.Status = Status_Dead;
		    /* Fall though */
		    break;
		default:
		    push_key(&event->key);
		    return key_exit;
		}
	    } else {
		if (sym == kbd_keys.fullscreen)
		    screen_togglefullscreen();
		else if (sym == kbd_keys.screenshot)
		    request_screenshot();
		else
		    switch (sym) {
		    case SDLK_PLUS:
		    case SDLK_EQUALS:
			screen_zoom(+1);
			break;
		    case SDLK_MINUS:
			screen_zoom(-1);
			break;
		    default:
			push_key(&event->key);
			return key_exit;
		    }
	    }
	    break;
	}
	default:
	    return 1;
	}
    }

    return 0;
}
